{#-
  This file was automatically generated - do not edit
-#}
{% extends "base.html" %}
{% block extrahead %}
  <link rel="stylesheet" href="{{ 'assets/stylesheets/custom.00c04c01.min.css' | url }}">
{% endblock %}
{% block scripts %}
  {{ super() }}
  <script src="{{ 'assets/javascripts/custom.129bd6ad.min.js' | url }}"></script>
{% endblock %}
